@@ -40,6 +40,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <termios.h>
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -57,13 +58,19 @@
 extern char *getpass (const char *__prompt);
 #endif
 
+#define BASE_LEN 21
+
 static long int server_port = 0;
 static char *cainfo = NULL;
 static char *cert = NULL;
 static char *sslkey = NULL;
 static char *endpoint = NULL;
 static char *username = NULL;
+static char *username_given = NULL; /* copy of either the username from env or cmdline*/
+static char *username_prev = NULL; /* input username to request_usr_pwd() when called last time */
 static char *password = NULL;
+static char *password_given = NULL; /* copy of either the password from env or cmdline */
+static char *password_prev = NULL; /* input password to request_usr_pwd() when called last time */
 static char *server = "localhost";
 static char *agent = NULL;
 static char *url_path = NULL;
@@ -103,6 +110,7 @@ static char *event_thumbprint = NULL;
 
 static char *cim_namespace = NULL;
 static char *fragment = NULL;
+static char *locale = NULL;
 static char *wsm_filter = NULL;
 static char *wsm_dialect = NULL;
 static char *input = NULL;
@@ -116,6 +124,7 @@ static char *output_file = NULL;
 static char *resource_uri_opt = NULL;
 static char *invoke_method = NULL;
 static char **properties = NULL;
+static char **option_set_values = NULL;
 
 struct _WsActions {
 	char *action;
@@ -222,6 +231,9 @@ static char wsman_parse_options(int argc, char **argv)
 		{"prop", 'k', U_OPTION_ARG_STRING_ARRAY, &properties,
 			"Properties with key value pairs (For 'put', 'invoke' and 'create')",
 			"<key=val>"},
+		{"option", (char)0, U_OPTION_ARG_STRING_ARRAY, &option_set_values,
+			"Option with key value pair for OptionSet (For 'put', 'invoke' and 'create')",
+			"<key=val>"},
 		{"config-file", 'C', U_OPTION_ARG_STRING, &config_file,
 			"Alternate configuration file", "<file>"},
 		{"out-file", 'O', U_OPTION_ARG_STRING, &output_file,
@@ -249,6 +261,8 @@ static char wsman_parse_options(int argc, char **argv)
 			"maximal envelope size", "<size>"},
 		{"fragment", 'F', U_OPTION_ARG_STRING, &fragment,
 			"Fragment (Supported Dialects: XPATH)", "<fragment>"},
+		{"locale", 'L', U_OPTION_ARG_STRING, &locale,
+			"Locale for this request", "<RFC 5646 language code>"},
 		{NULL}
 	};
 
@@ -373,6 +387,9 @@ static char wsman_parse_options(int argc, char **argv)
           }
           return FALSE;
 	}
+        else if (retval == 2) { /* help */
+          exit(0);
+        }
 
 	if (my_version) {
 		fprintf(stdout, PACKAGE_STRING " (" PACKAGE_BUILDTS ")\n\n");
@@ -488,49 +505,132 @@ request_usr_pwd( WsManClient *client, wsman_auth_type_t auth,
 	char *pw;
 	char user[21];
 	char *p;
+	char c;
+	int len = BASE_LEN;
+	int pos = 0;
+	struct termios term;
 
-	fprintf(stdout,"Authentication failed, please retry\n");
-	/*
-	   fprintf(stdout, "%s authentication is used\n",
-	   wsmc_transport_get_auth_name( auth));
-	   */
-	printf("User name: ");
-	fflush(stdout);
-	if ( (p = fgets(user, 20, stdin) ) != NULL )
-	{
+    /*
+	 * fprintf(stdout,"Authentication failed, please retry\n");
+     *
+	 * this message shall not be printed by this function as it cannot decide on the
+	 * reason it was called for. It does not control the authentication process.
+	 * wsmc_handler is better suited for such a decision making.
+	 */
 
-		if (strchr(user, '\n'))
-			(*(strchr(user, '\n'))) = '\0';
-		*username = u_strdup_printf ("%s", user);
-	} else {
-		*username = NULL;
+    if (username_given) {
+		if (password_given) {
+			/* Initially provided combination of password and username is not valid.
+			 * Request user to type both. Here I assume, that wsmc_handler called back to
+			 * this function after trying a first authentication using these credentials.
+			 */
+		} else {
+            /* Initially no password was provided => no authentication tried during first
+			 * iteration of while loop in wsmc_handler. Check previously typed credentials
+			 */
+			if (username_prev) {
+			   /* This is a second call of this function, assuming only wsmc_handler is using it
+			    * as a callback function. Therefore, there must have been a previous attempt to
+				* authenticate, but this previous combination of username and password did not
+				* lead to a successful authentication. Request new credentials, username_prev will
+				* be set each time after user has provided a username.
+				*/
+			} else {
+				/* First time wsmc_handler calls back to this function. No password given on the
+				 * command line or via the environment variable. Therefore wsmc_handler cannot
+				 * have tried http authentication. A username was given on the command line or
+				 * via an environment variable. And the user wants us to try this name at least
+				 * at first. So, let's do him a favour and use it. When we are called back again,
+				 * we will ask the user to provide a new name or the same, but a different password.
+				 */
+				*username = u_strdup(username_given);
+			}
+		}
 	}
 
-	pw = (char *)getpass("Password: ");
+    if (*username == NULL) {
+		printf("User name: ");
+		fflush(stdout);
+		if ( (p = fgets(user, 20, stdin) ) != NULL )
+		{
+			if (strchr(user, '\n'))
+				(*(strchr(user, '\n'))) = '\0';
+			*username = u_strdup_printf ("%s", user);
+		} else {
+			*username = NULL;
+		}
+	}
+
+	/* after successfull receipt of a new username, store a copy at username_prev */
+	if (*username) {
+		if ( username_prev ) {
+			u_free(username_prev);
+			username_prev = NULL;
+	}
+		username_prev = u_strdup(*username);
+	}
+
+    /* but always ask for the password !? */
+	/* set no echo */
+	tcgetattr(1, &term);
+	term.c_lflag &= ~ECHO;
+	tcsetattr(1, TCSANOW, &term);
+	/* get pass */
+	pw = (char*) u_malloc(sizeof(char) * len);
+	pw[0] = '\0';
+	printf("Password: ");
+	fflush(stdout);
+	while ((c = fgetc(stdin)) != '\n') {
+		pw[pos++] = (char) c;
+		if (pos >= len) {
+			len += BASE_LEN;
+			pw = realloc(pw, len);
+		}
+	}
+	pw[pos] = '\0';
 	*password = u_strdup_printf ("%s", pw);
+	u_free(pw);
+	/* set echo back again */
+	term.c_lflag |= ECHO;
+	tcsetattr(1, TCSANOW, &term);
+
+    /* make backup, *password will become free'd when next try of http-auth fails */
+    if (*password) {
+		if (password_prev) {
+			u_free(password_prev);
+			password_prev = NULL;
+		}
+		password_prev = u_strdup(*password);
+	}
 }
 
-
-
-static hash_t *wsman_options_get_properties(void)
+static void
+wsman_options_set_properties(client_opt_t *options)
 {
 	int c = 0;
-	hash_t *h = hash_create(HASHCOUNT_T_MAX, 0, 0);
 
 	while (properties != NULL && properties[c] != NULL) {
 		char *cc[3] = { NULL, NULL, NULL };
 		u_tokenize1(cc, 2, properties[c], '=');
-		if (!hash_lookup(h, cc[0])) {
-			if (!hash_alloc_insert(h, cc[0], cc[1])) {
-				debug("hash_alloc_insert failed");
-			}
-		} else {
-			warn("duplicate not added to hash");
-		}
-
+                wsmc_add_property(options, cc[0], cc[1]);
 		c++;
 	}
-	return h;
+	return;
+}
+
+
+static void
+wsman_options_set_option_set_values(client_opt_t *options)
+{
+	int c = 0;
+
+	while (option_set_values != NULL && option_set_values[c] != NULL) {
+		char *cc[3] = { NULL, NULL, NULL };
+		u_tokenize1(cc, 2, option_set_values[c], '=');
+                wsmc_add_option(options, cc[0], cc[1]);
+		c++;
+	}
+	return;
 }
 
 
@@ -648,6 +748,14 @@ int main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 
+    /* save copies of username or password when given on the command line or via environment variables */
+	if ( username != NULL ) {
+		username_given = u_strdup(username);
+	}
+	if ( password != NULL ) {
+		password_given = u_strdup(password);
+	}
+
 	filename = (char *) config_file;
 
 	if (filename) {
@@ -744,8 +852,13 @@ int main(int argc, char **argv)
 	if (fragment) {
 		options->fragment = fragment;
 	}
+	if (locale) {
+		options->locale = locale;
+	}
+
+	wsman_options_set_properties(options);
+	wsman_options_set_option_set_values(options);
 
-	options->properties = wsman_options_get_properties();
 	options->cim_ns = cim_namespace;
 	if (cim_extensions) {
 		wsmc_set_action_option(options, FLAG_CIM_EXTENSIONS);
@@ -884,7 +997,7 @@ int main(int argc, char **argv)
 				hnode_t *hn;
 				hash_t *selfilter = NULL;
 				hash_t *selectors_new = NULL;
-				selector_entry *entry;
+                                key_value_t *entry;
 				selectors_new = hash_create2(HASHCOUNT_T_MAX, 0, 0);
 				selfilter = u_parse_query(wsm_filter);
                                 if (!selfilter) {
@@ -893,9 +1006,9 @@ int main(int argc, char **argv)
                                 }
 				hash_scan_begin(&hs, selfilter);
 				while ((hn = hash_scan_next(&hs))) {
-					entry = u_malloc(sizeof(selector_entry));
+                                        entry = u_malloc(sizeof(key_value_t));
 					entry->type = 0;
-					entry->entry.text = (char *)hnode_get(hn);
+					entry->v.text = (char *)hnode_get(hn);
 					hash_alloc_insert(selectors_new, hnode_getkey(hn), entry);
 				}
 
